﻿<#@ template debug = "false" hostspecific = "false" language = "C#" #>
<#@ assembly name = "$(ProjectDir)bin\$rootnamespace$.dll" #>
<#@ assembly name = "System.Core.dll" #>
<#@ import namespace = "$rootnamespace$.Models" #>
<#@ import namespace = "System" #>
<#@ import namespace = "System.Collections" #>
<#@ import namespace = "System.Collections.Generic" #>
<#@ import namespace = "System.Linq" #>
<#@ import namespace = "System.Reflection" #>
<#@ import namespace = "System.Text" #>

<#@ output extension = ".d.ts" #>
///<reference path="../../Scripts/typings/knockout/knockout.d.ts" />
///<reference path="../../Scripts/typings/breeze/breeze.d.ts" />
interface EntityBase extends breeze.Entity {
	selected: KnockoutObservable<boolean>;
}
<#= TypeScriptConverter.GenerateTypeScriptDeclarations(true) #>

<#+

	public static class TypeScriptConverter
    {
        public static string GenerateTypeScriptDeclarations(bool useCamelCase)
        {
            var entityBaseType = typeof(EntityBase);

            var entityTypes = entityBaseType.Assembly.GetTypes().Where(t => t.IsSubclassOf(entityBaseType));
            var builder = new StringBuilder();

            return entityTypes
                .Aggregate(builder, (sb, t) => sb.Append(GenerateTypeScriptDeclaration(t, useCamelCase)))
                .ToString();
        }

        public static string GenerateTypeScriptDeclaration(Type modelType, bool useCamelCase)
        {

            var result = new StringBuilder();

            result.AppendFormat("interface {0} ", modelType.Name);

            if (modelType.BaseType != null
                && modelType.BaseType != typeof(object))
            {
                result.AppendFormat("extends {0} ", modelType.BaseType.Name);
            }

            result.Append("{");
            result.AppendLine();

            // Only get properties that are not derived
            var declaredProperties = modelType.GetProperties(
                BindingFlags.Public |
                BindingFlags.Instance |
                BindingFlags.DeclaredOnly
                );

            if (declaredProperties.Any())
            {

                foreach (var property in declaredProperties)
                {
					var prop = useCamelCase ? property.Name.Substring(0,1).ToLower() + property.Name.Substring(1) : property.Name;

                    var name = Nullable.GetUnderlyingType(property.PropertyType) == null
                        ? prop // not nullable
                        : (prop + "?"); // nullable

                    var type = ConvertTypeName(property.PropertyType);

                    result.AppendFormat("    {0} : {1};", name, type);
                    result.AppendLine();
                }
            }
            else
            {
                result.AppendLine("    ___Dummy : any;");
            }
            result.Append("}");
            result.AppendLine();

            result.AppendLine();

            //return Tuple.Create(GetReferencedTypes(modelType), result.ToString());
            return result.ToString();
        }

        private static string ConvertTypeName(Type type)
        {
            var family = GetTypeFamily(type);

            if (family == TypeFamily.SystemOrEnum)
            {
                var nullableUnderlyingType = Nullable.GetUnderlyingType(type);

                // Enums are integers
                if (type.IsEnum
                    || (nullableUnderlyingType != null
                        && nullableUnderlyingType.IsEnum))
                {
                    return "KnockoutObservable<number>";
                }

                var typeMappings = new Dictionary<string, string>
                {
                    { "System.Guid", "KnockoutObservable<string>" },
                    { "System.Int16", "KnockoutObservable<number>" },
                    { "System.Int32", "KnockoutObservable<number>" },
                    { "System.Int64", "KnockoutObservable<number>" },
                    { "System.UInt16", "KnockoutObservable<number>" },
                    { "System.UInt32", "KnockoutObservable<number>" },
                    { "System.UInt64", "KnockoutObservable<number>" },
                    { "System.Decimal", "KnockoutObservable<number>" },
                    { "System.Single", "KnockoutObservable<number>" },
                    { "System.Double", "KnockoutObservable<number>" },
                    { "System.Char", "KnockoutObservable<string>" },
                    { "System.String", "KnockoutObservable<string>" },
                    { "System.Boolean", "KnockoutObservable<boolean>" },
                    { "System.DateTime", "KnockoutObservable<Date>" }
                };

                var underlyingTypeName = type.ToString();

                // Check for nullables
                if (nullableUnderlyingType != null)
                {
                    underlyingTypeName = nullableUnderlyingType.ToString();
                }

                if (typeMappings.ContainsKey(underlyingTypeName))
                {
                    return typeMappings[underlyingTypeName];
                }

                return "any";
            }

            if (family == TypeFamily.Collection)
            {
                var elementType = type.GetGenericArguments().FirstOrDefault();

                return string.Format("KnockoutObservableArray<{0}>", elementType.Name);
            }

            // Single relationship to another model
            return string.Format("KnockoutObservable<{0}>", type.Name);
        }

        private static TypeFamily GetTypeFamily(Type type)
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }

            var isString = (type == typeof(string));
            var isEnumerable = typeof(IEnumerable).IsAssignableFrom(type);
            var isDictionary = type.FullName.StartsWith(typeof(IDictionary).FullName)
                || type.FullName.StartsWith(typeof(IDictionary<,>).FullName)
                || type.FullName.StartsWith(typeof(Dictionary<,>).FullName);

            if (!isString && !isDictionary && isEnumerable) return TypeFamily.Collection;

            if (type.Module.ScopeName == "CommonLanguageRuntimeLibrary" || type.IsEnum) return TypeFamily.SystemOrEnum;

            // Fallback when type is not recognised.
            return TypeFamily.Model;
        }

        private enum TypeFamily
        {
            SystemOrEnum,
            Model,
            Collection
        }
    }

#>